/*! ********************************************
 * GrayBox.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox;

public class GrayBox
{
	protected MidiEvent _eventList = null;
	protected ProcessData _processData = null;
	protected int _moduleSlots = 0;
	protected IGrayBoxModule[] _modules = null;
	protected boolean[] _bypassFlag = null;
	protected float _defAudioLatency = 0.0f;
	protected float _audioLatency    = 0.0f;
// Main System ---
	public float getSampleRate(){ return _processData.sampleRate; }
	public float setSampleRate(float sampleRate)
	{
		if (sampleRate < 1.0f)return _processData.sampleRate;
		_processData.sampleRate = sampleRate;
		for (int i = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].updateSampleRate(sampleRate);
		return _processData.sampleRate;
	}
	public int getAudioChannels(){ return _processData.audioChannels; }
	public int setAudioChannels(int ch)
	{
		if (ch < 2)return _processData.audioChannels;
		_processData.audioChannels = ch;
		for (int i = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].updateAudioChannels(ch);
		return _processData.audioChannels;
	}
	public int getProcessBlockSize(){ return _processData.blockSize; }
	public int setProcessBlockSize(int size)
	{
		if (size < 1)return _processData.blockSize;
		_processData.blockSize = size;
		for (int i = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].updateProcessBlockSize(size);
		return _processData.blockSize;
	}
	public float getDefaultAudioLatency(){ return _defAudioLatency; }
	public void setDefaultAudioLatency(float val) { _defAudioLatency = val; }
	public void audioLatency(float val) { _audioLatency = val; }
	public void initialize()
	{
		_defAudioLatency = 0.0f;
		_audioLatency = 0.0f;
		_clearProcessData();
		for (int i = 0; i < _moduleSlots; i++)
		{
			detachModule(i);
			_bypassFlag[i] = false;
		}
	}
// Module ---
	public int getModuleSlots() { return _moduleSlots; }
	public int setModuleSlots(int num)
	{
		if (num < 1 || num == _moduleSlots) return _moduleSlots;
		for (int i = 0; i < _moduleSlots; i++)detachModule(i);
		_modules = new IGrayBoxModule[num];
		_bypassFlag = new boolean[num];
		for (int i = 0; i < num; i++)
		{
			_modules[i] = null;
			_bypassFlag[i] = false;
		}
		return _moduleSlots = num;
	}
	public IGrayBoxModule getModule(int index)
	{
		if (index >= _moduleSlots) return null;
		return _modules[index];
	}
	public IGrayBoxModule attachModule(int index, IGrayBoxModule module)
	{
		if (index >= _moduleSlots) return null;
		if (module != null && module == _modules[index]) return module;
		for (int i = 0; i < _moduleSlots; i++)
		{
			if (module != _modules[i]) continue;
			IGrayBoxModule _mdl = _modules[i];
			_modules[i] = null;
			_mdl.disconnectModule();
		}
		_modules[index] = module;
		module.connectModule(this);
		return module;
	}
	public IGrayBoxModule detachModule(int index)
	{
		if (index >= _moduleSlots) return null;
		IGrayBoxModule _mdl = _modules[index];
		_modules[index] = null;
		return _mdl;
	}
	public boolean getBypassStatus(int index)
	{
		if (index >= _moduleSlots) return false;
		return _bypassFlag[index];
	}
	public void bypassProcess(int index, boolean bypass)
	{
		if (index >= _moduleSlots) return;
		_bypassFlag[index] = bypass;
		if (_modules[index] != null)_modules[index].updateBypassStatus(bypass);
	}
	public void resetModules()
	{
		for (int i = 0; i < _moduleSlots; i++) if (_modules[i] != null)_modules[i].resetModule();
	}
	public void initiModules()
	{
		for (int i = 0; i < _moduleSlots; i++) if (_modules[i] != null)_modules[i].initiModule();
	}
// Process ---
	public void suspendProcess()
	{
		for (int i = 0; i < _moduleSlots; i++) if (_modules[i] != null)_modules[i].suspendProcess();
	}
	public void resumeProcess()
	{
		for (int i = 0; i < _moduleSlots; i++) if (_modules[i] != null)_modules[i].resumeProcess();
	}
	public void startProcess()
	{
		for (int i = 0; i < _moduleSlots; i++) if (_modules[i] != null)_modules[i].startProcess();
	}
	public void stopProcess()
	{
		for (int i = 0; i < _moduleSlots; i++) if (_modules[i] != null)_modules[i].stopProcess();
	}
	protected void _clearProcessData()
	{
		_processData.midiEvents = null;
		_processData.audioOut = null;
		_processData.offsetFrames = 0;
		_processData.sampleFrames = 0;
		_audioLatency = _defAudioLatency;
	}
	public void process(float[][] audioOut, int offsetFrames, int sampleFrames)
	{
	// begin Process -->
		_eventList.clear();
		_processData.midiEvents = _eventList;
		_processData.audioOut = audioOut;
		_processData.offsetFrames = offsetFrames;
		_processData.sampleFrames = sampleFrames;
		_processData.audioLatency = _audioLatency;
	// Modules Process -->
		for (int i = 0; i < _moduleSlots; i++) if (_modules[i] != null && !_bypassFlag[i]) _modules[i].process(_processData);
	// after Process -->
		_clearProcessData();
	}
// Constructor
	public GrayBox(int blockSize, float sampleRate){ this(blockSize, sampleRate, 2, 4); }
	public GrayBox(int blockSize, float sampleRate, int audioChannels){ this(blockSize, sampleRate, audioChannels, 4); }
	public GrayBox(int blockSize, float sampleRate, int audioChannels, int moduleSlots)
	{
		// processData
		_eventList = new MidiEvent();
		_processData = new ProcessData();
		_processData.blockSize = (blockSize < 1)? 1 : blockSize;
		_processData.sampleRate = (sampleRate < 1.0f)? 1.0f : sampleRate;
		_processData.audioChannels = (audioChannels >= 2)? audioChannels : 2;
		// module
		_moduleSlots = (moduleSlots < 1)? 1 : moduleSlots;
		_modules = new IGrayBoxModule[_moduleSlots];
		_bypassFlag = new boolean[_moduleSlots];
		for (int i = 0; i < _moduleSlots; i++)
		{
			_modules[i] = null;
			_bypassFlag[i] = false;
		}
	}
}
