/*! ********************************************
 * EFXStereoDelay.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.efx;
import sora.graybox.GrayBox;
import sora.graybox.ProcessData;
import sora.graybox.IGrayBoxModule;
public class EFXStereoDelay extends EffectorBase implements IGrayBoxModule
{
// EFXStereoDelay --- 
	// Ring Buffer
	private int _ringBufferSizeBits = 15;
	private int _ringBufferSize = 1 << _ringBufferSizeBits;
	private int _ringBufferMask = _ringBufferSize - 1;
	private float[] _delayBufferL;
	private float[] _delayBufferR;
	// 
	private float _sampleRate = 44100.0f;
	// Delay
	private int _wp = 0;  // WritePoint [frame]
	// Delay L
	private float _feedBackL  = 0.4f; // 0 ~ 1.0
	private int _rpL          = 0;      // ReadPoint [frame]
	private int _delayL       = 0;      // [frame]
	private float _delayTimeL = 0.0f; // [Sec]
	// Delay R
	private float _feedBackR  = 0.4f;
	private int _rpR          = 0;
	private int _delayR       = 0;
	private float _delayTimeR = 0.0f;
// getter, setter
	public int getRingBufferSize() { return _ringBufferSize; }
	public float getSampleRate() { return _sampleRate; }
	public void setSampleRate(float rate)
	{
		_sampleRate = rate;
		_delayTimeL = _delayL / _sampleRate;
		_delayTimeR = _delayR / _sampleRate;
	}
	public float getDelayFeedBackLevelL() { return _feedBackL; }
	public float getDelayFeedBackLevelR() { return _feedBackR; }
	public void setDelayFeedBackLevelL(float val) { _feedBackL = val; }
	public void setDelayFeedBackLevelR(float val) { _feedBackR = val; }
	
	public float getDelayTimeL() { return _delayTimeL; }
	public float getDelayTimeR() { return _delayTimeR; }
	public float setDelayTimeL(float val)
	{
		int _frame = (int)(val * _sampleRate);
		if (_frame > _ringBufferMask)_frame = _ringBufferMask;
		_delayL = _frame;
		_rpL = _wp + (_ringBufferMask - _delayL) & _ringBufferMask;
		return _delayTimeL = _frame / _sampleRate;
	}
	public float setDelayTimeR(float val)
	{
		int _frame = (int)(val * _sampleRate);
		if (_frame > _ringBufferMask)_frame = _ringBufferMask;
		_delayR = _frame;
		_rpR = _wp + (_ringBufferMask - _delayR) & _ringBufferMask;
		return _delayTimeR = _frame / _sampleRate;
	}
	
	public int getDelayFrameL() { return _delayL; }
	public int getDelayFrameR() { return _delayR; }
	public void setDelayFrameL(int val)
	{ 
		_delayL = val;
		if (_delayL > _ringBufferMask)_delayL = _ringBufferMask;
		_rpL = _wp + (_ringBufferMask - _delayL) & _ringBufferMask;
		_delayTimeL = _delayL / _sampleRate;
	}
	public void setDelayFrameR(int val)
	{
		_delayR = val;
		if (_delayR > _ringBufferMask)_delayR = _ringBufferMask;
		_rpR = _wp + (_ringBufferMask - _delayR) & _ringBufferMask;
		_delayTimeR = _delayR / _sampleRate;
	}
	public void setEffectParam(){ setEffectParam( 0.2f, 0.25f, 0.4f, 0.35f, 0.4f); }
	public void setEffectParam(float mix, float delayTimeL, float feedBackLevelL, float delayTimeR, float feedBackLevelR)
	{
		_mix = mix;
		_antiMix = 1.0f - _mix;
		setDelayTimeL(delayTimeL);
		setDelayTimeR(delayTimeR);
		_feedBackL = feedBackLevelL;
		_feedBackR = feedBackLevelR;
	}
// Constructor
	public EFXStereoDelay(){ this(15); }
	public EFXStereoDelay(int bufferSizeBits)
	{
		super();
		// bufferSize
		if (bufferSizeBits > 20) bufferSizeBits = 20;
		else if (bufferSizeBits < 4) bufferSizeBits = 4;
		_ringBufferSizeBits = bufferSizeBits;
		_ringBufferSize = 1 << _ringBufferSizeBits;
		_ringBufferMask = _ringBufferSize - 1;
		_delayBufferL = new float[_ringBufferSize];
		_delayBufferR = new float[_ringBufferSize];
		// reset
		clearBuffer();
		setEffectParam();
	}
// Base - EffectorBase ---
	@Override
	public void applyEffect(float[] outL, float[] outR, int startFrame, int sampleFrames)
	{
		int i = startFrame, _endFrame = i + sampleFrames;
		for ( ; i < _endFrame; i++,
			_rpL = _rpL + 1 & _ringBufferMask,
			_rpR = _rpR + 1 & _ringBufferMask,
			_wp  = _wp  + 1 & _ringBufferMask)
		{
		// Channel L --->
			float _delay = _delayBufferL[_rpL];
			_delayBufferL[_wp] = outL[i] + _delay * _feedBackL;
			outL[i] = outL[i] * _antiMix + _delay * _mix;	// outout
		// Channel R --->
			_delay = _delayBufferR[_rpR];
			_delayBufferR[_wp] = outR[i] + _delay * _feedBackR;
			outR[i] = outR[i] * _antiMix + _delay * _mix;	// outout
		}
	}
	@Override
	public void resetParameters()
	{
		setEffectParam();
	}
	@Override
	public void clearBuffer()
	{
		for (int i = 0; i < _ringBufferSize; i++)_delayBufferL[i] = _delayBufferR[i] = 0.0f;
	}
	@Override
	public void initialize()
	{
		clearBuffer();
		setEffectParam();
	}
// Interface - IfGrayBoxModule -----
	public void updateSampleRate(float sampleRate){ setSampleRate(sampleRate); }
	public void updateAudioChannels(int ch){}
	public void updateProcessBlockSize(int size){}
	
	public void resetModule(){ initialize(); }
	public void initiModule(){ initialize(); }
	public void connectModule(GrayBox host){ clearBuffer(); setSampleRate(host.getSampleRate()); }
	public void disconnectModule(){ clearBuffer(); }
	public void updateBypassStatus(boolean bypass){ clearBuffer(); }
	
	public void startProcess(){ clearBuffer(); }
	public void stopProcess(){ clearBuffer(); }
	public void suspendProcess(){ clearBuffer(); }
	public void resumeProcess(){ clearBuffer(); }
	public void process(ProcessData procData)
	{
		// inline --- applyEffect()
		float[] _outL = procData.audioOut[0], _outR = procData.audioOut[1];
		int i = procData.offsetFrames, _endFrame = i + procData.sampleFrames;
		for ( ; i < _endFrame; i++,
			_rpL = _rpL + 1 & _ringBufferMask,
			_rpR = _rpR + 1 & _ringBufferMask,
			_wp  = _wp  + 1 & _ringBufferMask)
		{
		// Channel L ---------
			float _delay = _delayBufferL[_rpL];
			_delayBufferL[_wp] = _outL[i] + _delay * _feedBackL;
			_outL[i] = _outL[i] * _antiMix + _delay * _mix;	// outout
		// Channel R ---------
			_delay = _delayBufferR[_rpR];
			_delayBufferR[_wp] = _outR[i] + _delay * _feedBackR;
			_outR[i] = _outR[i] * _antiMix + _delay * _mix;	// outout
		}
	}
}
