/*! ********************************************
 * EFXStereoReverb.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.efx;
import sora.graybox.GrayBox;
import sora.graybox.ProcessData;
import sora.graybox.IGrayBoxModule;
public class EFXStereoReverb extends EffectorBase implements IGrayBoxModule
{
// EFXStereoReverb --- 
	// Ring Buffer
	private int _ringBufferSizeBits = 15;
	private int _ringBufferSize = 1 << _ringBufferSizeBits;
	private int _ringBufferMask = _ringBufferSize - 1;
	private float[] _delayBuffer1;
	private float[] _delayBuffer2;
	// 
	private float _sampleRate = 44100.0f;
	// Delay
	private int _wp = 0;  // WritePoint [frame]
	// Delay 1
	private float _feedBack1  = 0.4f; // 0 ~ 1.0
	private int _rp1          = 0;      // ReadPoint [frame]
	private int _delay1       = 0;      // [frame]
	private float _delayTime1 = 0.0f; // [Sec]
	// Delay 2
	private float _feedBack2  = 0.4f;
	private int _rp2          = 0;
	private int _delay2       = 0;
	private float _delayTime2 = 0.0f;
// getter, setter
	public int getRingBufferSize() { return _ringBufferSize; }
	public float getSampleRate() { return _sampleRate; }
	public void setSampleRate(float rate)
	{
		_sampleRate = rate;
		_delayTime1 = _delay1 / _sampleRate;
		_delayTime2 = _delay2 / _sampleRate;
	}
	public float getDelayFeedBackLevel1() { return _feedBack1; }
	public float getDelayFeedBackLevel2() { return _feedBack2; }
	public void setDelayFeedBackLevel1(float val) { _feedBack1 = val; }
	public void setDelayFeedBackLevel2(float val) { _feedBack2 = val; }
	
	public float getDelayTime1() { return _delayTime1; }
	public float getDelayTime2() { return _delayTime2; }
	public float setDelayTime1(float val)
	{
		int _frame = (int)(val * _sampleRate);
		if (_frame > _ringBufferMask)_frame = _ringBufferMask;
		_delay1 = _frame;
		_rp1 = _wp + (_ringBufferMask - _delay1) & _ringBufferMask;
		return _delayTime1 = _frame / _sampleRate;
	}
	public float setDelayTime2(float val)
	{
		int _frame = (int)(val * _sampleRate);
		if (_frame > _ringBufferMask)_frame = _ringBufferMask;
		_delay2 = _frame;
		_rp2 = _wp + (_ringBufferMask - _delay2) & _ringBufferMask;
		return _delayTime2 = _frame / _sampleRate;
	}
	public int getDelayFrame1() { return _delay1; }
	public int getDelayFrame2() { return _delay2; }
	public void setDelayFrame1(int val)
	{ 
		_delay1 = val;
		if (_delay1 > _ringBufferMask)_delay1 = _ringBufferMask;
		_rp1 = _wp + (_ringBufferMask - _delay1) & _ringBufferMask;
		_delayTime1 = _delay1 / _sampleRate;
	}
	public void setDelayFrame2(int val)
	{
		_delay2 = val;
		if (_delay2 > _ringBufferMask)_delay2 = _ringBufferMask;
		_rp2 = _wp + (_ringBufferMask - _delay2) & _ringBufferMask;
		_delayTime2 = _delay2 / _sampleRate;
	}
	public void setEffectParam(){ setEffectParam( 0.2f, 0.25f, 0.4f, 0.35f, 0.4f); }
	public void setEffectParam(float mix, float delayTime1, float feedBackLevel1, float delayTime2, float feedBackLevel2)
	{
		_mix = mix;
		_antiMix = 1.0f - _mix;
		setDelayTime1(delayTime1);
		setDelayTime2(delayTime2);
		_feedBack1 = feedBackLevel1;
		_feedBack2 = feedBackLevel2;
	}
// Constructor
	public EFXStereoReverb(){ this(15); }
	public EFXStereoReverb(int bufferSizeBits)
	{
		super();
		// bufferSize
		if (bufferSizeBits > 20) bufferSizeBits = 20;
		else if (bufferSizeBits < 4) bufferSizeBits = 4;
		_ringBufferSizeBits = bufferSizeBits;
		_ringBufferSize = 1 << _ringBufferSizeBits;
		_ringBufferMask = _ringBufferSize - 1;
		_delayBuffer1 = new float[_ringBufferSize];
		_delayBuffer2 = new float[_ringBufferSize];
		// reset
		clearBuffer();
		setEffectParam();
	}
// Base - EffectorBase ---
	@Override
	public void applyEffect(float[] outL, float[] outR, int startFrame, int sampleFrames)
	{
		int i = startFrame, _endFrame = i + sampleFrames;
		for ( ; i < _endFrame; i++,
			_rp1 = _rp1 + 1 & _ringBufferMask,
			_rp2 = _rp2 + 1 & _ringBufferMask,
			_wp  = _wp  + 1 & _ringBufferMask)
		{
		// Channel L --->
			float _delay = _delayBuffer1[_rp1];
			_delayBuffer1[_wp] = outL[i] + _delay * _feedBack1;
			outL[i] = outL[i] * _antiMix + _delay * _mix;	// outout
		// Channel R --->
			_delay = _delayBuffer2[_rp2];
			_delayBuffer2[_wp] = outR[i] + _delay * _feedBack2;
			outR[i] = outR[i] * _antiMix + _delay * _mix;	// outout
		}
	}
	@Override
	public void resetParameters()
	{
		setEffectParam();
	}
	@Override
	public void clearBuffer()
	{
		for (int i = 0; i < _ringBufferSize; i++)_delayBuffer1[i] = _delayBuffer2[i] = 0.0f;
	}
	@Override
	public void initialize()
	{
		clearBuffer();
		setEffectParam();
	}
// Interface - IfGrayBoxModule -----
	public void updateSampleRate(float sampleRate){ setSampleRate(sampleRate); }
	public void updateAudioChannels(int ch){}
	public void updateProcessBlockSize(int size){}
	
	public void resetModule(){ initialize(); }
	public void initiModule(){ initialize(); }
	public void connectModule(GrayBox host){ clearBuffer(); setSampleRate(host.getSampleRate()); }
	public void disconnectModule(){ clearBuffer(); }
	public void updateBypassStatus(boolean bypass){ clearBuffer(); }
	
	public void startProcess(){ clearBuffer(); }
	public void stopProcess(){ clearBuffer(); }
	public void suspendProcess(){ clearBuffer(); }
	public void resumeProcess(){ clearBuffer(); }
	public void process(ProcessData procData)
	{
		// inline --- applyEffect()
		float[] _outL = procData.audioOut[0], _outR = procData.audioOut[1];
		int i = procData.offsetFrames, _endFrame = i + procData.sampleFrames;
		for ( ; i < _endFrame; i++,
			_rp1 = _rp1 + 1 & _ringBufferMask,
			_rp2 = _rp2 + 1 & _ringBufferMask,
			_wp  = _wp  + 1 & _ringBufferMask)
		{
		// Channel L ---------
			float _delay = _delayBuffer1[_rp1];
			_delayBuffer1[_wp] = _outL[i] + _delay * _feedBack1;
			_outL[i] = _outL[i] * _antiMix + _delay * _mix;	// outout
		// Channel R ---------
			_delay = _delayBuffer2[_rp2];
			_delayBuffer2[_wp] = _outR[i] + _delay * _feedBack2;
			_outR[i] = _outR[i] * _antiMix + _delay * _mix;	// outout
		}
	}
}
